// -*- C++ -*-
// $RCSfile: automatrix.h,v $
// $Revision: 1.16 $
// $Author: reida $
// $Date: 2006-03-03 22:07:00 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

// Create a reference counted matrix class from a base matrix class.
// If the base class is a matrix M of T's, it must have the following:
//	M::M(int, int);         // initialize to given size
//	M::M(const T&);		// copy constructor
//    	typedef T value_type;
//    	int nrows() const;
//	int ncols() const;
//	T& operator()(int, int);
//	const T& operator()(int, int) const;
//	std::ostream& operator<<(std::ostream&, const M&);

#ifndef AUTOMATRIX_H
#define AUTOMATRIX_H

#include <iostream>
#include "engine/matvec.h"
#include "engine/matrixbase.h"
template <class TYPE>
class AutoMatrix : public MatrixBase {
private:
  int *counter;
  TYPE *mat;
  bool owner;

  // make a copy of mat, unless this is the only reference to it.
  TYPE *conditional_copy() {
    if(*counter > 1) {
      --*counter;		// we no longer refer to the old mat
      counter = new int;
      *counter = 1;
      mat = new TYPE(*mat);	// use TYPE's copy constructor
    }
    return mat;
  }

  // delete the underlying matrix and counter if there are no references left.
  void cleanup() {
    //std::cerr << "cleanup " << mat << " (" << *counter << ")" << std::endl;
    if(*counter == 0) {
      delete counter;
      if(owner)
	delete mat;
    }
  }

public:
  // Create an AutoMatrix from a new TYPE matrix of the specified size.
  AutoMatrix(int n, int m)
    : mat(new TYPE(n, m)),
      counter(new int),
      owner(true)
  {
    *counter = 1;
  }

  // Create an AutoMatrix from an existing TYPE matrix.  The TYPE
  // matrix must have been allocated with "new".  If owner==true, the
  // AutoMatrix takes over ownership, and will delete the matrix when
  // appropriate.
  AutoMatrix(TYPE *m, bool owner)
    : counter(new int),
      mat(m),
      owner(owner)
  {
    *counter = 1;
  }

  // the copy constructor doesn't actually make a copy, it just makes
  // a new reference.
  AutoMatrix(const AutoMatrix<TYPE> &am)
    : counter(am.counter),
      mat(am.mat),
      owner(false)
  {
    ++*counter;
  }

  // delete the underlying matrix only if there are no more references
  // to it.
  virtual ~AutoMatrix() {
    --*counter;
    cleanup();
  }

  // assignment operator
  AutoMatrix<TYPE> &operator=(const AutoMatrix<TYPE> &other) {
    // remove reference to current matrix
    --*counter;
    cleanup();
    // now refer to other matrix
    mat = other.mat;
    counter = other.counter;
    ++*counter;
    return *this;
  }

  // implicit conversion to the underlying TYPE matrix
  operator const TYPE&() const { return *mat; }
  // If implicit conversion is ambiguous, use underlying_matrix() to
  // make it explicit.
  const TYPE &underlying_matrix() const { return *mat; }

  int nrows() const { return mat->nrows(); }
  int ncols() const { return mat->ncols(); }
  int nnz() const { return mat->nnz(); }

  // non-const access requires that we make a copy of the matrix, if
  // there is more than one reference to it.
  typename TYPE::value_type &operator()(int i, int j) {
    // std::cerr << "&operator()(int i, int j) 1" << std::endl;
    return conditional_copy()->operator()(i, j);
    
  }

  typedef typename TYPE::const_iterator const_iterator;
//   typename TYPE::const_iterator begin() const { 
//     return TYPE::const_iterator(*mat);
//       // ...or even mat->begin();
//   }

  //typename TYPE::const_iterator end() const {
  //  return mat->end();
  // }

  // const access can be safely done without making a copy.
  /*
  const typename TYPE::value_type operator()(int i, int j) const {
    //std::cerr << "operator()(int i, int j) 2" << std::endl;
    return mat->operator()(i, j);
  }

  */

  friend std::ostream &operator<<(std::ostream &os, const AutoMatrix<TYPE> &am)
  {
    os << *(am.mat);
    return os;
  }

  template <class VEC>
  VEC operator*(const VEC &v) const {
    return (*mat)*v;
  }

  template <class VEC>
  VEC trans_mult(const VEC &v) const {
    return mat->trans_mult(v);
  }

  AutoMatrix<TYPE>& operator*=(const typename TYPE::value_type &x) {
    conditional_copy()->operator*=(x);
    return *this;
  }
  virtual VECTOR_D operator*(const VECTOR_D &x) const {
    return (*mat)*(x);
    
};
  virtual MatrixBase &operator*=(double x){
    conditional_copy()->operator*=(x);
    return *this;
  }
  virtual VECTOR_D trans_mult(const VECTOR_D &v) const {
    return mat->trans_mult(v);
  }
  
  virtual double operator() (int i, int j) const {
    return (*mat)(i,j);
    
  }
  
//   virtual AutoMatrix<TYPE>  add(const TYPE &b) const
//   {
//     return AutoMatrix<TYPE>(mat->add(b), true);
//   }

//   virtual AutoMatrix<TYPE> &operator+=(const AutoMatrix<TYPE> &b) {
//     mat += b.mat;
//     return *this;
//   }
  
};

// template<class TYPE>
// AutoMatrix<TYPE> operator+(const AutoMatrix<TYPE> &a, const AutoMatrix<TYPE> &b)
//   {
// //     return a.add(b);
//     AutoMatrix<TYPE> result(a);
//     result += b;
//     return result;
//   }

#endif
