// -*- C++ -*-
// $RCSfile: flux.C,v $
// $Revision: 1.63 $
// $Author: langer $
// $Date: 2007-03-23 20:12:03 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/threadstate.h"
#include "common/trace.h"
#include "engine/csubproblem.h"
#include "engine/edge.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/fluxnormal.h"
#include "engine/indextypes.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/mvmult.h"
#include "engine/nodalequation.h"
#include "engine/ooferror.h"
#include "engine/outputval.h"
#include "engine/symmmatrix.h"

const std::string Flux::modulename_("oof2.SWIG.engine.flux");

std::vector<Flux*> &Flux::allfluxes() {
  static std::vector<Flux*> all_fluxes;
  return all_fluxes;
}

Flux *getFluxByIndex(int index) {
  return Flux::allfluxes()[index];
}

int countFluxes() {
  return Flux::allfluxes().size();
}

Flux *Flux::getFlux(const std::string &name) {
  const std::vector<Flux*> &list = allfluxes();
  for(std::vector<Flux*>::size_type i=0; i<list.size(); i++) {
    if(list[i]->name() == name) {
      return list[i];
    }
  }
  throw ErrProgrammingError("Unknown Flux \"" + name + "\"",
			    __FILE__, __LINE__);
}

Flux::Flux(const std::string &nm, int d, int dvdim)
  : name_(nm),
    dim(d),
    divdim(dvdim)
{
  index_ = allfluxes().size();
  allfluxes().push_back(this);
}

bool operator==(const Flux &fluxa, const Flux &fluxb) {
  return fluxa.index_ == fluxb.index_;
}

bool operator!=(const Flux &fluxa, const Flux &fluxb) {
  return fluxa.index_ != fluxb.index_;
}

std::ostream &operator<<(std::ostream &os, const Flux &flux) {
  os << "Flux(" << flux.name() << ")";
  return os;
}

void Flux::addEquation(Equation *eqn) {
  // Add eqn to the list of equations in which this flux appears, but
  // first make sure that it's not already in the list.
  for(std::vector<Equation*>::size_type i=0; i<eqnlist.size(); i++) {
    if(eqnlist[i] == eqn) return;
  }
  eqnlist.push_back(eqn);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

const std::string &VectorFlux::classname() const {
  static std::string name_("VectorFlux");
  return name_;
}

const std::string &SymmetricTensorFlux::classname() const {
  static std::string name_("SymmetricTensorFlux");
  return name_;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

IteratorP VectorFlux::iterator(Planarity planarity) const {
  int maxdim = 3;
  int mindim = 0;
  if(planarity == IN_PLANE) maxdim = 2;
  if(planarity == OUT_OF_PLANE) mindim = 2;
  return IteratorP(new VectorFieldIterator(mindim, maxdim));
}

IteratorP VectorFlux::divergence_iterator() const {
  return IteratorP(new ScalarFieldIterator);
}

IteratorP VectorFlux::out_of_plane_iterator() const {
  return IteratorP(new OutOfPlaneVectorFieldIterator());
}

IndexP VectorFlux::componenttype() const {
  return IndexP(new VectorFieldIndex);
}

IndexP VectorFlux::getIndex(const std::string &str) const {
  return IndexP(new VectorFieldIndex(str[0] - 'x'));
}

IndexP VectorFlux::divergence_componenttype() const {
  return IndexP(new ScalarFieldIndex);
}

IndexP VectorFlux::divergence_getIndex(const std::string&) const {
  return IndexP(new ScalarFieldIndex);
}

IteratorP SymmetricTensorFlux::iterator(Planarity planarity) const {
  if(planarity == IN_PLANE)
    return IteratorP(new SymTensorInPlaneIterator);
  if(planarity == OUT_OF_PLANE)
    return IteratorP(new SymTensorOutOfPlaneIterator);
  return IteratorP(new SymTensorIterator);
}

IteratorP SymmetricTensorFlux::divergence_iterator() const
{
  return IteratorP(new VectorFieldIterator(0, 2));
}

IteratorP SymmetricTensorFlux::out_of_plane_iterator() const {
  return IteratorP(new OutOfPlaneSymTensorIterator());
}

IndexP SymmetricTensorFlux::componenttype() const {
  return IndexP(new SymTensorIndex);
}

IndexP SymmetricTensorFlux::getIndex(const std::string &str) const {
  return IndexP(new SymTensorIndex(SymTensorIndex::str2voigt(str)));
}

IndexP SymmetricTensorFlux::divergence_componenttype() const {
  return IndexP(new VectorFieldIndex);
}

IndexP SymmetricTensorFlux::divergence_getIndex(const std::string &str) const {
  return IndexP(new VectorFieldIndex(str[0] - 'x'));
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

FluxData *Flux::initializeData(const Element *el) const {
  return new FluxData(this, el);
}

FluxData::FluxData(const Flux *flux, const Element *el) {
  // The flux matrix is the matrix that multiplies the vector of dofs
  // of an element to get the flux at a point.  Each property makes
  // its contribution to the flux matrix by calling
  // FluxData::matrix_element().
  fluxmatrix_ = new MATRIX_D(flux->dim, el->ndof(), 0.0);
  fluxoffset_ = new VECTOR_D(flux->dim, 0.0);

  // The right-hand-side contribution occurs in the form of the
  // *divergence* of the flux, that is to say, as a force, so it has
  // the dimensionality "divdim".  It gets specified at a point by the
  // property's "fluxrhs" routine, and then integrated and distributed
  // amongst the nodes.
  fluxrhs_ = new VECTOR_D(flux->divdim, 0.0);
}

FluxData::~FluxData() {
  delete fluxmatrix_;
  delete fluxoffset_;
  delete fluxrhs_;
}

double &FluxData::matrix_element(const FieldIndex &fluxindex,
				 const Field *field,
				 const FieldIndex &fieldindex,
				 const ElementFuncNodeIterator &node)
{
  int row = fluxindex.integer();
  int col = node.localindex(*field, &fieldindex);
  return (*fluxmatrix_)(row, col);
}

double FluxData::matrix_element(const FieldIndex &fluxindex,
				const Field *field,
				const FieldIndex &fieldindex,
				const ElementFuncNodeIterator &node) const
{
  int row = fluxindex.integer();
  int col = node.localindex(*field, &fieldindex);
  return (*fluxmatrix_)(row, col);
}

// Special versions for scalar fields, where the field component
// shouldn't have to be specified.

double &FluxData::matrix_element(const FieldIndex &fluxindex,
				 const Field *field,
				 const ElementFuncNodeIterator &node)
{
  ScalarFieldIndex fieldindex;
  int row = fluxindex.integer();
  int col = node.localindex(*field, &fieldindex);
  return (*fluxmatrix_)(row, col);
}

double FluxData::matrix_element(const FieldIndex &fluxindex,
				const Field *field,
				const ElementFuncNodeIterator &node) const
{
  ScalarFieldIndex fieldindex;
  int row = fluxindex.integer();
  int col = node.localindex(*field, &fieldindex);
  return (*fluxmatrix_)(row, col);
}

// Indexing functions for the offset vector.  Require only one index,
// of course, and comes in const and non-const versions.
double FluxData::offset_element(const FieldIndex &fluxindex) const {
  return (*fluxoffset_)[fluxindex.integer()];
}

double &FluxData::offset_element(const FieldIndex &fluxindex) {
  return (*fluxoffset_)[fluxindex.integer()];
}

// Analogous indexing functions for the rhs.  The fluxrhs is, during
// the integration, the value of the body force (corresponding to the
// divergence of the current flux) at a point.

double FluxData::rhs_element(int idx) const {
  return (*fluxrhs_)[idx];
}

double &FluxData::rhs_element(int idx) {
  return (*fluxrhs_)[idx];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

VECTOR_D *Flux::evaluate(const FEMesh *mesh, const Element *el,
			 const MasterPosition &pos) const
{
  VECTOR_D *result = new VECTOR_D(dim, 0.0);
  if(el->material()!=NULL) {
    FluxData *fluxdata = initializeData(el);
    el->material()->find_fluxmatrix(mesh, el, this, fluxdata, pos);
    VECTOR_D doflist(el->localDoFs());
    *result = (*fluxdata->fluxmatrix())*doflist + *fluxdata->fluxoffset();
    delete fluxdata;
  }
  return result;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Unlike "evaluate", this returns a vector (not a pointer)

// VECTOR_D Flux::evaluate_flux(const CSubProblem *subp, const Element *el,
// 			     const MasterPosition &pos) const
// {
//   FluxData *fluxdata = initializeData(el);
//   el->material()->find_fluxmatrix(subp, el, this, fluxdata, pos);
//   VECTOR_D doflist(el->localDoFs(subp));
//   VECTOR_D result(dim, 0.0);
//   result = (*fluxdata->fluxmatrix())*doflist + *fluxdata->fluxoffset();
//   delete fluxdata;
//   return result;
// }

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Flux::make_stiffness(CSubProblem *subp, const Element *el,
			  const GaussPoint &gpt,
			  const FluxData *fluxdata,
			  LinearizedSystem &system)
{
  // Loop over all active Equations that use this Flux
  for(std::vector<Equation*>::size_type i=0; i<eqnlist.size(); i++) {
    Equation *eqn = eqnlist[i];
    // TODO PLASTICITY: this may not be the correct check:
    if(subp->is_active_equation(*eqn))
      eqn->make_stiffness(subp, el, gpt, *this, fluxdata, system);
  }
}

// Analogue of make_stiffness.  Called by Material::make_linear_system
// while integrating over gauss points within an element.

void Flux::make_rhs(CSubProblem *subp, const Element *el, const GaussPoint &gpt,
		    const FluxData *fluxdata, LinearizedSystem &system)
{
  for(std::vector<Equation*>::size_type i=0; i<eqnlist.size(); i++) {
    Equation *eqn = eqnlist[i];
    if(subp->is_active_equation(*eqn)) {
      eqn->make_rhs(subp, el, gpt, *this, fluxdata, system);
    }
  }
}

int Flux::eqn_integration_order(const CSubProblem *subp, const Element *el)
  const
{
  int order = 0;
  for(std::vector<Equation*>::size_type i=0; i<eqnlist.size(); ++i) {
    Equation *eqn = eqnlist[i];
    if(subp->is_active_equation(*eqn)) {
      int ord = eqn->integration_order(el);
      if(ord > order)
	order = ord;
    }
  }
  return order;
}

// Flux boundary condition uses the same mechanism as the stiffness
// matrix, only it makes its contribution to the boundary right-hand-side,
// rather than to the global stiffness matrix.
void Flux::boundary_integral(CSubProblem *subp, const BoundaryEdge *ed, 
			     const EdgeGaussPoint &egp,
			     const FluxNormal *flxnormal) const {
  for(std::vector<Equation*>::size_type i=0; i<eqnlist.size(); i++) {
    Equation *eq = eqnlist[i];
    if(subp->is_active_equation(*eq)) {
      eq->boundary_integral(subp, this, ed, egp, flxnormal);
    }
  }
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Assuming that the flux matrix has been calculated, compute the
// matrix which, when integrated over the element and multiplied by
// the vector of dofs, gives the divergence of the flux (up to
// boundary terms which come from integration by parts).  The rows of
// the matrix are indexed by the component of the divergence and the
// element's node number.  This is the local stiffness matrix for
// divergence equations.

// This routine computes the contributions to this matrix for a single
// node nu.

void SymmetricTensorFlux::local_stiffness(const Element *el,
					  const ElementFuncNodeIterator &nu,
					  const GaussPoint &gpt,
					  const FluxData *fluxdata,
					  MATRIX_D &matrix) const
{
  // The flux matrix is M_{(ij) (\mu l)}, where (ij) is the flux component.
  // The local stiffness matrix is
  //    K_{(i \nu) (\mu l)} = (sign) \nabla_j N_\nu M_{(ij) (\mu l)}
  // the (sign) is positive, unless the symmetry and positive definiteness
  // of the matrix can be preserved for some of the fields.
  // Please see heatconductivity.py for an example.

  // Assume that matrix is initialized already.

  const int ncols = el->ndof();

  // Tabulate shapefunction derivatives. Even though they're cached,
  // there are a lot of lookups and virtual functions required.
  std::vector<double> dsf(spacedim);
  for(SpaceIndex j=0; j<spacedim; ++j)
    dsf[j] = nu.dshapefunction(j, gpt);

  for(SpaceIndex i=0; i<spacedim; ++i) {
    // i only goes up to spacedim, even when spacedim is 2 and the
    // flux is a 3x3 symmetric tensor.  This is because the
    // out-of-plane component of the flux divergence equation isn't
    // being solved.  It can be computed a posteriori.
    for(int dof=0; dof<ncols; dof++) {
      double sum = 0;
      for(SpaceIndex j=0; j<spacedim; ++j) {
	SymTensorIndex ij(i, j);
	double kij = dsf[j]*(*fluxdata->fluxmatrix())(ij.integer(), dof);
	sum += kij;
      }
      matrix(i, dof) += sum;
    }
  }
}


// Local_boundary function is the analogue of the local_stiffness
// function, except that it's much simpler.  The core assumption here
// is that the components of the equation will match up with
// components of the flux normal in the obvious way, so that the sizes
// will match and the assignment will be straightforward.
void SymmetricTensorFlux::local_boundary(const BoundaryEdge *ed, 
					 EdgeNodeIterator& edi,
					 const EdgeGaussPoint &egp, 
					 const FluxNormal *flxnormal,
					 std::vector<double>& rhs) const {

  const SymTensorFluxNormal *flxn =
    dynamic_cast<const SymTensorFluxNormal*>(flxnormal);
  // Check size of rhs -- has to be 2 for this flux.
  int rhssize = rhs.size();
  if (rhssize == SYMTEN_DIV_DIM) {
    if (flxn->size() == rhssize) { // Simple case, do the obvious.
      double sfvalue = edi.shapefunction(egp);
      rhs[0] = sfvalue * flxn->x;
      rhs[1] = sfvalue * flxn->y;
    }
  }
}

void SymmetricTensorFlux::make_offset(const Element *el, 
				      const ElementFuncNodeIterator &mu,
				      const GaussPoint &gpt,
				      const FluxData *fluxdata,
				      VECTOR_D &offset_rhs) const {

  std::vector<double> dsf(spacedim);
  // Make contributions to the right hand side from the flux offset vector.
  // Tabulate shape-function derivatives, following the example of 
  // local_stiffness. 
  for(SpaceIndex i=0;i<spacedim;++i) {
    dsf[i] = mu.dshapefunction(i,gpt);
    offset_rhs[i] = 0.0;  // Also clear the vector.
  }

  // Actual computation is a simple index contraction over i.
  // Dimensions match, because spacedim is equal to divdim for
  // a symmetric tensor flux.  
  for(SpaceIndex i=0; i<spacedim; ++i) {
    for(SpaceIndex j=0; j<spacedim; ++j) {
      SymTensorIndex ij(i,j);
      offset_rhs[j] += (*fluxdata->fluxoffset())[ij.integer()]*dsf[i];
    }
  }
}

// Evaluate the dot product of the flux with the normal at
// the indicated edge gauss point.
VECTOR_D *SymmetricTensorFlux::contract(const FEMesh *mesh,
					const Element *elmt,
					const EdgeGaussPoint &egpt) const {
  VECTOR_D *value = evaluate(mesh, elmt, egpt);
  // voigt order, 00, 11, 22, 12, 02, 01
  VECTOR_D *result = new VECTOR_D(divdim, 0.0);
  Coord normal = egpt.normal();
  (*result)[0] = (*value)[0]*normal(0) + (*value)[5]*normal(1);
  (*result)[1] = (*value)[5]*normal(0) + (*value)[1]*normal(1);
  delete value;  // deleting new'd VECTOR_D from "evaluate".
  return result;
}

FluxNormal *SymmetricTensorFlux::BCCallback(const Coord &pos,
					    double time, 
					    const Coord &nrm,
					    const double distance,
					    const double fraction,
					    PyObject *wrapper,
					    const PyObject *pyfunction)
  const {

  PyObject *args;
  PyObject *result;
  Coord cres;

  acquirePyLock();
  args = Py_BuildValue("(Oddddddd)", pyfunction, pos(0), pos(1), time,
		       nrm(0), nrm(1), distance, fraction);
  result = PyEval_CallObject(wrapper, args);
  Py_DECREF(args);
  if(result) {
    if(PyTuple_Check(result)) {
      if(PyTuple_Size(result) == (Py_ssize_t) 2) {
	cres(0) = PyFloat_AsDouble(PyTuple_GET_ITEM(result, (Py_ssize_t) 0));
	cres(1) = PyFloat_AsDouble(PyTuple_GET_ITEM(result, (Py_ssize_t) 1));
      }
      else {
	// Only one "release" per possible control-flow path.
	releasePyLock();
	throw 
	  ErrSetupError(
		       "SymmetricTensorFlux::BCCallback: Wrong size of tuple.");
      }
    }   
    else {
      releasePyLock();
      throw ErrSetupError("SymmetricTensorFlux::BCCallback: Expected a tuple.");
    }
  }
  else {			// !result.  PyEval_CallObject failed.
    releasePyLock();
    pythonErrorRelay();
  }
  releasePyLock();
  Py_XDECREF(result);
  return new SymTensorFluxNormal(cres(0),cres(1));
}

OutputValue Flux::output(const FEMesh *mesh, const Element *el,
			 const MasterPosition &pos) const
{
  VECTOR_D *fluxvals = evaluate(mesh, el, pos);
  OutputValue ov = newOutputValue();
  for(IteratorP it = iterator(ALL_INDICES); !it.end(); ++it) {
    ov[it] = (*fluxvals)[it.integer()];
  }
  return ov;
}

OutputValue SymmetricTensorFlux::newOutputValue() const {
  SymmMatrix3 *sm3 = new SymmMatrix3();
//   std::cerr << "SymmetricTensorFlux::newOutputValue: " << sm3 << std::endl;
  return OutputValue(sm3);
//   return OutputValue(new SymmMatrix3());
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

OutputValue VectorFlux::newOutputValue() const {
  return OutputValue(new VectorOutputVal(ndof()));
}

void VectorFlux::local_stiffness(const Element *el,
				 const ElementFuncNodeIterator &nu,
				 const GaussPoint &gpt,
				 const FluxData *fluxdata,
				 MATRIX_D &matrix) const
{
  // The flux matrix is M_{i\mu}, where i is the flux component.
  // The local stiffness matrix is
  //    K_{\nu\mu} = -\nabla_i N_\nu M_{i\mu}

  const int ncols = el->ndof();

  // Tabulate shapefunction derivatives. Even though they're cached,
  // there are a lot of lookups and virtual functions required.
  std::vector<double> dsf(spacedim);
  for(SpaceIndex i=0; i<spacedim; ++i)
    dsf[i] = nu.dshapefunction(i, gpt);

  for(int dof=0; dof<ncols; dof++) {
    double sum = 0;
    for(SpaceIndex i=0; i<spacedim; ++i) {
      sum += dsf[i] * (*fluxdata->fluxmatrix())(i, dof);
    }
    matrix(0, dof) -= sum; //minus sign accounts for flux going down
			   //hill!! and guarantees
			   //positive-definiteness
  }
}

// The VectorFlux boundary condition calls the correct wrapper
// for its callback, and assumes the components match up
// trivially, in the absence of indications otherwise from 
// the equation.
void VectorFlux::local_boundary(const BoundaryEdge *ed,
				EdgeNodeIterator& edi,
				const EdgeGaussPoint &egp,
				const FluxNormal *flxnormal,
				std::vector<double>& rhs) const {
  // Size.
  const VectorFluxNormal *flxn = 
    dynamic_cast<const VectorFluxNormal*>(flxnormal);
  int rhssize = rhs.size();
  if (rhssize == VEC_DIV_DIM) {
    if (flxn->size() == rhssize) {
      rhs[0] = edi.shapefunction(egp) * flxn->value();
    }
  }
}

void VectorFlux::make_offset(const Element *el, 
			     const ElementFuncNodeIterator &mu, 
			     const GaussPoint &gpt, 
			     const FluxData *fluxdata,
			     VECTOR_D &offsetrhs) const {

  std::vector<double> dsf(spacedim);
  for(SpaceIndex i=0;i<spacedim;++i) {
    dsf[i] = mu.dshapefunction(i,gpt);
  }
  offsetrhs[0] = 0.0;   // Only one component of vector divergence!

  // Index contraction is very simple in this case.
  for(SpaceIndex i=0;i<spacedim;++i) {
    offsetrhs[0] += (*fluxdata->fluxoffset())[i]*dsf[i];
  }
}

// Evalute the dot product of the flux with the normal at the
// indicated gauss point.
VECTOR_D *VectorFlux::contract(const FEMesh *mesh,
			       const Element *elmt,
			       const EdgeGaussPoint &egpt) const {
  VECTOR_D *value = evaluate(mesh, elmt, egpt);
  Coord normal = egpt.normal();
  VECTOR_D *result = new VECTOR_D(divdim, 0.0);
  (*result)[0] = (*value)[0]*normal(0) + (*value)[1]*normal(1);
  return result;
}


FluxNormal *VectorFlux::BCCallback(const Coord &pos,
				   double time, 
				   const Coord &nrm,
				   const double distance, 
				   const double fraction,
				   PyObject *wrapper,
				   const PyObject *pyfunction) const {
  PyObject *args;
  PyObject *result;
  double dres = 0.0;

  acquirePyLock();
  args = Py_BuildValue("(Oddddddd)",pyfunction, pos(0), pos(1), time, 
		       nrm(0), nrm(1), distance, fraction);
  result = PyEval_CallObject(wrapper, args);
  Py_DECREF(args);

  if(result) {
    if(PyTuple_Check(result)) {
      if(PyTuple_Size(result)==1) {
	dres = PyFloat_AsDouble(PyTuple_GET_ITEM(result, (Py_ssize_t) 0));
      }
      else {
	// Only one "release" per possible control-flow path.
	releasePyLock();
	throw 
	  ErrSetupError("VectorFlux::BCCallback: Wrong size of tuple.");
      }
    }   
    else {
      releasePyLock();
      throw ErrSetupError("VectorFlux::BCCallback: Expected a tuple.");
    }
  }

  Py_XDECREF(result);
  releasePyLock();

  return new VectorFluxNormal(dres);
}
