// -*- C++ -*-
// $RCSfile: solverdriver.C,v $
// $Revision: 1.21 $
// $Author: langer $
// $Date: 2007-09-13 21:34:04 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modifed
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "engine/csrmatrix.h"
#include "engine/csubproblem.h"
#include "engine/matvec.h"
#include "engine/solver.h"
#include "engine/solverdriver.h"
#include "engine/steppingscheme.h"

double dot_product(VECTOR_D &left, VECTOR_D &right)
{
  double result = 0;
  for (VECTOR_D::size_type i = 0 ; i<left.size(); i++)
    result += left(i)*right(i);
  return result;

}

SolverDriverCore::~SolverDriverCore() {
  //std::cerr << "destroying SolverDriverCore object" << std::endl;
  //clear_memory();
  //std::cerr << "leaving SolverDriverCore destroy call" << std::endl;
}

SolverDriverCore::SolverDriverCore(SteppingScheme *stepsch,
				   LinearSolver *solver, CSubProblem *subp) 
  : solver(solver),
    subproblem(subp),
    sscheme(stepsch),
    K(0),
    b(0),
    nrowsA(0), 
    ncolsA(0),
    A(0),
    C(0),
    x(0),
    rhs(0),
    xp(0)
{
}

void SolverDriverCore::set_solver_components(LinearSolver *ls,
					     CSubProblem *subp)
{
  solver = ls;
  subproblem = subp;
}
double SolverDriverCore::residual()
{
  //same residual as the one used by the SparseLib solvers
  return solver->get_residual();
  
}

void SolverDriverCore::precompute() {
  //std::cerr << "SolverDriverCore::precomputing"  << std::endl;
  //at this point the solver and the preconditioner are already selected
  // at the UI and passed as arguments into the solver
  RWLock * rwlock = subproblem->mesh->get_rwlock();
  rwlock->read_acquire(); //acquire read lock
  //std::cerr << "after the lock" << std::endl;
  try{
    int n, m;
    if (A) {
      delete A;
    }
    A = new CSRmatrix (subproblem->Amatrix()); 
    
    nrowsA = n = A->nrows();
    
//     std::cerr << "SolverDriverCore::precompute: " << A->nrows() << "x" << A->ncols() << std::endl;
    ncolsA = A->ncols();
    if(n != A->ncols())
      throw ErrSetupError("Stiffness matrix is not square! nrows="
			  + to_string(n) + ", ncols="
			  + to_string(A->ncols()));
    
    if (C) {
      delete C;
    }
    C = new CSRmatrix (subproblem->Cmatrix());
    m = C->ncols();
    if (!x) {
      x = new VECTOR_D (n, 0.0);
    }
    x->clear();
    subproblem->mesh->get_unknowns(*x, subproblem->freedofmap);
    //std::cerr << "x=" << *x << std::endl;
    if (!rhs)
    {
      ////std::cerr << "allocating memory for rhs" << std::endl;
      rhs = new VECTOR_D (n, 0.0);
    }
    rhs->clear();
    subproblem->get_rhs(*rhs, subproblem->indepeqnmap);
    
    if (!xp) {
      //std::cerr << "allocating memory for xp" << std::endl;
      xp = new VECTOR_D (m, 0.0);
    }
    xp->clear();
    //std::cerr << "allocating memory for xp" << std::endl;
    subproblem->mesh->get_unknowns(*xp, subproblem->fixeddofmap);
    *rhs -= (*C) * (*xp);
    
    
  } //end of try block
  catch(...) //guarantee that lock will be released if ANY exception is raised
    {
      //std::cerr << "releasing the lock in catch" << std::endl;
      rwlock->read_release(); 
      throw;
    }
  //std::cerr << "releasing the lock while exiting the function " << std::endl;
  rwlock->read_release();

  //stepper interactions
  //these lines should be outside the locked area
  sscheme->precompute(subproblem, A, rhs, x);
						
  // Get the K matrix and b vector of the linear system, and store it.
  // We don't actually own these matrices, but they're stored as
  // object attributes so they're common to precompute, solve, and
  // postcompute.
  //  std::cerr << "A: " << std::endl;
  // std::cerr << *A << std::endl;
  K = sscheme->get_K(subproblem, A);
  b = sscheme->get_b(subproblem, A, rhs);
  // std::cerr << "K: " << std::endl;
  // std::cerr << *K << std::endl;
  // std::cerr << "b:" << std::endl;
  // std::cerr << *b << std::endl;
}


bool SolverDriverCore::solve(){
  //std::cerr << "entering solverStepperCore::solve" << std::endl;

  // TODO: Does set_blocking really have to be called all the time? It
  // seems to be required only by one preconditioner.  Can't the
  // preconditioner request it if necessary?
  solver->set_blocking(subproblem);
  
  solver->precompute(*K);
  
  bool status = solver->solve(*K, *b, *x);

  solver->postcompute(); // Call possibly-null post-compute hook.
  //postcompute();
  //std::cerr << "leaving solverStepperCore::solve" << std::endl;
  return status;
  
}

void SolverDriverCore::postcompute() {
  //std::cerr << "SolverDriverCore::postcomputing"  << std::endl;
  RWLock * rwlock = subproblem->mesh->get_rwlock();
  rwlock->write_acquire();
  try {
     subproblem->mesh->set_unknowns(*x, subproblem->freedofmap);

     // In the old days, when nodal equations had values, the values
     // of the dependent equations were computed here.  There's no
     // actual need to do this, but here's how it was done:
     // CSRmatrix Cp = femesh->Cpmatrix();
     // CSRmatrix B = femesh->Bmatrix();
     // femesh->set_rhs(Cp * (*x) + B * (*xp), femesh->depeqnmap);

     sscheme->set_x(x); //stepping scheme creates a copy of solution
  }
  catch(...) {
    rwlock->write_release();
  }
  rwlock->write_release();
  sscheme->postcompute(subproblem);
}

void SolverDriverCore::clear_memory() {
  //std::cerr << "entering SolverDriverCore::clear_memory()" << std::endl;
  delete A;
  //std::cerr << "after deleting A" << std::endl;
  delete C;
  //std::cerr << "after deleting C" << std::endl;
  delete x;
  //std::cerr << "after deleting x" << std::endl;
  delete rhs;
  //std::cerr << "after deleting rhs" << std::endl;
  delete xp;
  //std::cerr << "after deleting xp" << std::endl;
  sscheme->clean_up();
  K=0;
  b=0;
  //std::cerr << "leaving SolverDriverCore::clear_memory()" << std::endl;
}

NewtonDriverCore::NewtonDriverCore(SteppingScheme *stepsch,
				   LinearSolver *solver, CSubProblem *subp)
  : SolverDriverCore(stepsch, solver, subp),
    psi(0),
    dx(0)
{}

NewtonDriverCore::~NewtonDriverCore()
{
  delete psi; //free memory of psi
  delete dx;
}

double NewtonDriverCore::residual()
{
  //same residual as the one used by the SparseLib solvers
  if (!psi || !b) {
    return 0;
  }
  double psi_mag = dot_product(*psi, *psi);
  double rhs_mag = dot_product(*b, *b);
  if (rhs_mag == 0.0) 
    rhs_mag = 1;
  return sqrt(psi_mag/rhs_mag);
}

void NewtonDriverCore::precompute()
{
  SolverDriverCore::precompute();
  RWLock * rwlock = subproblem->mesh->get_rwlock();
  rwlock->read_acquire();
  try {
    if (!psi) {
      // at this point the fields have been initialized, so it is okay
      // to build psi
      psi = new VECTOR_D (K->nrows(), 0.0);
    }
    // TODO: Bug?  Psi is later used as the RHS of the linear system
    // which is solved to determine dx.  This should be the vector of
    // values of the equations evaluated at the new x, which is not
    // given by the below expression -- we can compute this directly,
    // instead of using this approximate form.
    *psi = (*b) - (*K) * (*x); 
    if (dx)
      delete dx;
    dx = new VECTOR_D (K->nrows(), 0.0);
  }
  catch(...) {
    rwlock->read_release(); 
      throw;
  }
  rwlock->read_release(); 
}


bool NewtonDriverCore::solve() {
  // Take a single step of the NR method. Use the linear solver to
  // solve for the increment, dx.  psi is the residual, b-Ax.
  solver->set_blocking(subproblem);
  solver->precompute(*K);	// preconditioning, if any
  bool status = solver->solve(*K, *psi, *dx);
  if (status) 
    (*x) += (*dx);		// Apply the computed increment.
  solver->postcompute(); 
  return status;
}
